"""
This is library and CLI utils for controlling RV6688BCM router
It is required python 3.5 or higher and requests library (due to Digest HTTP auth)

The MIT License (MIT)
Copyright (c) 2016 Baryshnikov Alexander <dev@baryshnikov.net>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from enum import Enum
from typing import List
from rvcm.cli import *


class ForwardType(Enum):
    """
    Enumeration of possible protocol types in forwarding rules
    """
    TCP = 1
    UDP = 2
    BOTH = 3

    def __str__(self):
        return self.name


class Forward:
    """
    Describes port forwarding single rule in the router
    """

    def __init__(self, name='', dest_ip_sec='', src_min_port=0, src_max_port=0, dest_min_port=0, dest_max_port=0,
                 enabled=False,
                 type: ForwardType = ForwardType.BOTH):
        self.dest_ip_sec = dest_ip_sec
        self.src_min_port = src_min_port
        self.src_max_port = src_max_port
        self.dest_min_port = dest_min_port
        self.dest_max_port = dest_max_port
        self.name = name
        self.enabled = enabled
        self.type = type

    def __repr__(self):
        return self.__class__.__name__ + "(" + ", ".join(k + "=" + repr(v) for k, v in self.__dict__.items()) + ")"

    def __str__(self):
        return "%s-%s-%s-%s-%s-%s-%s-%s-0-" % (
            1 if self.enabled else 0,
            self.name,
            self.src_min_port,
            self.src_max_port,
            self.type.value,
            self.dest_min_port,
            self.dest_max_port,
            self.dest_ip_sec
        )


class NAT:
    """
    Describes table of port forwarding rules in the router
    """
    URL = '/vs.htm?l0=1&l1=2&l2=0&l3=-1'
    UPDATE = '/setup.cgi?l0=1&l1=2&l2=0&l3=-1'

    def __init__(self, forwards: List[Forward] = None):
        self.forwards = forwards or []

    def parse(self, data):
        """
        Parse content of NAT page and gather info into self structure
        :param data: Content of NAT page info
        """
        fws = []
        for line in data.splitlines():
            if line.startswith('var vs_list'):
                fws = line[15:-3].split(';')
                break
        self.forwards.clear()
        for rule in fws:
            opt = rule.split('-')
            forward = Forward()
            forward.enabled = int(opt[0]) != 0
            forward.name = opt[1]
            forward.src_min_port = int(opt[2])
            forward.src_max_port = int(opt[3])
            forward.type = ForwardType(int(opt[4]))
            forward.dest_min_port = int(opt[5])
            forward.dest_max_port = int(opt[6])
            forward.dest_ip_sec = int(opt[7])
            self.forwards.append(forward)

    def retrieve(self, requester: callable):
        """
        Get information about NAT from router
        :param requester: function that returns text by url
        :return: self
        """
        resp = requester(self.URL)
        self.parse(resp.text)
        return self

    def save(self, poster: callable):
        """
        Save information about NAT in router
        :param poster: function that post data to router by URL
        """
        from collections import OrderedDict
        params = OrderedDict()
        for key, value in self.generate_form_fields():
            params[key] = value
        poster(self.UPDATE, params, self.URL)

    def generate_form_fields(self):
        """
        Generate pairs of forms fields
        :return: iterator of (key, value) tuple
        """
        yield ('virtual_server_list', 'Active Worlds')
        yield ('vs_pc_list', '0')
        yield ('if_list', '')
        yield ('clear_entry_list', '')
        for i, forward in enumerate(self.forwards):
            yield ('enable_%s' % i, str(1 if forward.enabled else 0))
            yield ('description_%s' % i, forward.name)
            yield ('inbound_port_low_%s' % i, str(forward.src_min_port))
            yield ('inbound_port_high_%s' % i, str(forward.src_max_port))
            yield ('type_%s' % i, str(forward.type.value))
            yield ('private_port_low_%s' % i, str(forward.dest_min_port))
            yield ('private_port_high_%s' % i, str(forward.dest_max_port))
            yield ('private_ip_%s' % i, str(forward.dest_ip_sec))
            yield ('if_%s' % i, '0')
        yield ('h_vs_list', ";".join(str(f) for f in self.forwards) + ";")
        yield ('fwi_des', '')
        yield ('todo', 'save')
        yield ('this_file', 'vs.htm')
        yield ('next_file', 'vs.htm')
        yield ('message', '')

    def __repr__(self):
        return self.__class__.__name__ + "(" + ",\n   ".join(k + "=" + repr(v) for k, v in self.__dict__.items()) + ")"

    def __str__(self):
        return repr(self)


@cli.group()
def nat():
    """
    NAT operations
    """
    pass


@nat.command()
@click.pass_context
def info(ctx):
    """
    Print forwarding table
    """
    nat = NAT().retrieve(ctx.obj.getter)
    preip = ctx.obj.ip.split('.')[:-1]
    print("{name:21s} {state:10s} {srcmin:10s} {srcmax:10s} {dstmin:10s} {dstmax:10s} {proto:5s} {ip:4s}".format(
        name="NAME",
        state="STATUS",
        srcmin="S-MIN-PRT",
        srcmax="S-MAX-PRT",
        dstmin="D-MIN-PRT",
        dstmax="D-MAX-PRT",
        proto="PROTO",
        ip="IP"

    ))
    for forward in nat.forwards:
        line = "{name:21s} {state:10s} {srcmin:10s} {srcmax:10s} {dstmin:10s} {dstmax:10s} {proto:5s} {ip:4s}".format(
            name=forward.name,
            state='active' if forward.enabled else 'inactive',
            srcmin=str(forward.src_min_port),
            srcmax=str(forward.src_max_port),
            proto=str(forward.type.name),
            dstmin=str(forward.dest_min_port),
            dstmax=str(forward.dest_max_port),
            ip=".".join(preip + [str(forward.dest_ip_sec)])
        )
        print(line)


@nat.command()
@click.argument('name')
@click.argument('dest-ip-section')
@click.argument('min-src-port')
@click.argument('max-src-port')
@click.argument('min-dst-port')
@click.argument('max-dst-port')
@click.option('--proto', default=ForwardType.BOTH.name,
              type=click.Choice([ForwardType.BOTH.name, ForwardType.TCP.name, ForwardType.UDP.name]),
              help='Protocol type to forward')
@click.pass_context
def create(ctx, name: str, dest_ip_section: str, min_src_port: int, max_src_port: int, min_dst_port: int, max_dst_port,

           proto: ForwardType):
    """
    Create forwarding rule
    """
    assert 0 < len(dest_ip_section) <= 3, "IP last section required"

    forward = Forward(name=name, dest_ip_sec=dest_ip_section)
    forward.src_min_port = min_src_port
    forward.src_max_port = max_src_port
    forward.dest_min_port = min_dst_port
    forward.dest_max_port = max_dst_port
    forward.enabled = False
    forward.type = ForwardType[proto]
    nat = NAT().retrieve(ctx.obj.getter)
    nat.forwards.append(forward)
    nat.save(ctx.obj.poster)


@nat.command()
@click.argument('name')
@click.option('--dest-ip-section', type=int)
@click.option('--min-src-port', type=int)
@click.option('--max-src-port', type=int)
@click.option('--min-dst-port', type=int)
@click.option('--max-dst-port', type=int)
@click.option('--proto', default=ForwardType.BOTH.name,
              type=click.Choice([ForwardType.BOTH.name, ForwardType.TCP.name, ForwardType.UDP.name]),
              help='Protocol type to forward')
@click.pass_context
def update(ctx, name: str, dest_ip_section: str, min_src_port: int, max_src_port: int, min_dst_port: int, max_dst_port,
           proto: ForwardType):
    """Update forwarding record"""
    upd = False
    nat = NAT().retrieve(ctx.obj.getter)
    for frw in nat.forwards:
        if frw.name == name:
            if max_dst_port is not None and frw.dest_max_port != max_dst_port:
                frw.dest_max_port = max_dst_port
                upd = True
            if min_dst_port is not None and frw.dest_min_port != min_dst_port:
                frw.dest_min_port = min_dst_port
                upd = True
            if min_src_port is not None and frw.src_min_port != min_src_port:
                frw.src_min_port = min_src_port
                upd = True
            if max_src_port is not None and frw.src_max_port != max_src_port:
                frw.src_max_port = max_src_port
                upd = True
            if dest_ip_section is not None and frw.dest_ip_sec != dest_ip_section:
                frw.dest_ip_sec = dest_ip_section
                upd = True
            if proto is not None and frw.type != proto:
                frw.type = ForwardType[proto]
                upd = True
            if upd:
                print("updating " + str(frw))

    if upd:
        nat.save(ctx.obj.poster)
    else:
        print("Nothing to update")


@nat.command()
@click.argument('name')
@click.pass_context
def remove(ctx, name: str):
    """Remove forwarding rule"""
    upd = False
    nat = NAT().retrieve(ctx.obj.getter)
    items = []
    for frw in nat.forwards:
        if frw.name != name:
            items.append(frw)
        else:
            upd = True
            print("removing " + str(frw))
    nat.forwards = items
    if upd:
        nat.save(ctx.obj.poster)
    else:
        print("nothing to remove")


@nat.command()
@click.argument('name')
@click.pass_context
def disable(ctx, name: str):
    """Disable (but not remove) rule"""
    nat = NAT().retrieve(ctx.obj.getter)
    upd = False
    for frw in nat.forwards:
        if frw.name == name and frw.enabled:
            frw.enabled = False
            upd = True
            print("disabling " + str(frw))
    if upd:
        nat.save(ctx.obj.poster)
    else:
        print("nothing to disable")


@nat.command()
@click.argument('name')
@click.pass_context
def enable(ctx, name: str):
    """Enable rule"""
    nat = NAT().retrieve(ctx.obj.getter)
    upd = False
    for frw in nat.forwards:
        if frw.name == name and not frw.enabled:
            frw.enabled = True
            upd = True
            print("enabling " + str(frw))
    if upd:
        nat.save(ctx.obj.poster)
    else:
        print("nothing to enable")


@nat.command()
@click.argument('old_name')
@click.argument('new_name')
@click.pass_context
def rename(ctx, old_name, new_name):
    """Rename forwarding rule"""
    nat = NAT().retrieve(ctx.obj.getter)
    upd = False
    for frw in nat.forwards:
        if frw.name == old_name:
            print("renaming " + str(frw))
            upd = True
            frw.name = new_name
    if upd:
        nat.save(ctx.obj.poster)
    else:
        print("nothing to rename")


if __name__ == '__main__':
    cli(obj=None)



"""
This is library and CLI utils for controlling RV6688BCM router
It is required python 3.5 or higher and requests library (due to Digest HTTP auth)

The MIT License (MIT)
Copyright (c) 2016 Baryshnikov Alexander <dev@baryshnikov.net>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from rvcm.cli import *
from lxml import html
import json


class Info:
    """
    Describes full information about router status
    """
    # URL to page with full information
    URL = '/index.htm'

    def __init__(self, ip='', gateway='', mac='', sip_user='', local_ip='', dns1='', dns2='', firmware='', model='',
                 gpon_serial='', phone_line_up=False, wan_line_up=False, lan_line_up=False, apply_required=False):
        self.ip = ip
        self.gateway = gateway
        self.mac = mac
        self.sip_user = sip_user
        self.local_ip = local_ip
        self.dns1 = dns1
        self.dns2 = dns2
        self.firmware = firmware
        self.model = model
        self.gpon_serial = gpon_serial
        self.phone_line_up = phone_line_up
        self.wan_line_up = wan_line_up
        self.lan_line_up = lan_line_up
        self.apply_required = apply_required

    def parse(self, page: str):
        """
        Parse index HTML page and gather info
        :param page: Content of index page
        """
        root = html.fromstring(page)
        form = root.find('.//form[@action="setup.cgi"]')

        self.wan_line_up = '"Up"' in form.find('.//table/tr[2]/td/table/tr[2]/td[4]/script').text
        self.ip = form.find('.//table/tr[2]/td/table/tr[3]/td[2]').text
        self.gateway = form.find('.//table/tr[2]/td/table/tr[3]/td[4]').text
        self.dns1 = form.find('.//table/tr[2]/td/table/tr[4]/td[2]').text
        self.dns2 = form.find('.//table/tr[2]/td/table/tr[4]/td[4]').text

        self.phone_line_up = '"Up"' in form.find('.//table/tr[4]/td/table/tr[3]/td[2]/script').text
        self.sip_user = form.find('.//table/tr[4]/td/table/tr[4]/td[2]').text

        self.model = form.find('.//table/tr[6]/td/table/tr[2]/td[2]').text
        self.firmware = form.find('.//table/tr[6]/td/table/tr[2]/td[4]').text
        self.gpon_serial = form.find('.//table/tr[6]/td/table/tr[3]/td[2]').text
        self.mac = form.find('.//table/tr[6]/td/table/tr[3]/td[4]').text

        self.local_ip = form.find('.//table/tr[8]/td/table/tr[2]/td[2]').text
        self.lan_line_up = '"Connected"' in form.find('.//table/tr[8]/td/table/tr[2]/td[4]/script').text

        self.apply_required = False
        for line in page.splitlines():
            if line == 'var headMsg = "Please do Apply to make the changes take effect.";':
                self.apply_required = True
                break

    def retrieve(self, requester: callable):
        """
        Get information about NAT from router
        :param requester: function that returns text by url
        :return: self
        """
        resp = requester(self.URL)
        self.parse(resp.text)
        return self

    def pretty(self):
        """
        Make pretty-printed text with router info
        :return: text
        """

        lines = []
        lines += ["Model           : {}".format(self.model)]
        lines += ["IP              : {}".format(self.ip)]
        lines += ["Gateway         : {}".format(self.gateway)]
        lines += ["MAC             : {}".format(self.mac)]
        lines += ["Local IP        : {}".format(self.local_ip)]
        lines += ["DNS1            : {}".format(self.dns1)]
        lines += ["DNS2            : {}".format(self.dns2)]
        lines += ["SIP             : {}".format(self.sip_user)]
        lines += ["GPON            : {} ".format(self.gpon_serial)]
        lines += ["Firmware        : {} ".format(self.firmware)]
        lines += ["Phone status    : {}".format("OK" if self.phone_line_up else "Unavailable")]
        lines += ["WAN status      : {}".format("OK" if self.wan_line_up else "Unavailable")]
        lines += ["LAN status      : {}".format("OK" if self.lan_line_up else "Unavailable")]
        lines += ["Unsaved changes : {}".format("Yes" if self.apply_required else "No")]
        return "\n".join(lines)

    def __repr__(self):
        return self.__class__.__name__ + "(" + ",\n   ".join(k + "=" + repr(v) for k, v in self.__dict__.items()) + ")"

    def __str__(self):
        return repr(self)


@cli.group()
def router():
    """
    Direct router operations
    """
    pass


@router.command()
@click.pass_context
def apply(ctx):
    """
    Apply changes on the router
    """
    ctx.obj.poster('/setup.cgi?l0=-1&l1=-1&l2=-1&l3=-1', {
        'todo': 'apply',
        'this_file': 'global_nav.htm',
        'next_file': 'index.html'
    })


@router.command()
@click.pass_context
def info(ctx):
    """Print details about router"""
    info = Info().retrieve(ctx.obj.getter)
    print(info.pretty())


@router.command()
@click.pass_context
def export(ctx):
    """Print details about router in json"""
    info = Info().retrieve(ctx.obj.getter)
    data = {
        "model": info.model,
        "ip": info.ip,
        "gateway": info.gateway,
        "mac": info.mac,
        "local_ip": info.local_ip,
        "dns1": info.dns1,
        "dns2": info.dns2,
        "sip": info.sip_user,
        "gpon": info.gpon_serial,
        "firmware": info.firmware,
        "phone_status": info.phone_line_up,
        "wan_status": info.wan_line_up,
        "lan_status": info.lan_line_up,
        "unsaved_changes": info.apply_required,
    }
    print(json.dumps(data, ensure_ascii=False, indent=4))


if __name__ == '__main__':
    cli(obj=None)



"""
This is library and CLI utils for controlling RV6688BCM router
It is required python 3.5 or higher and requests library (due to Digest HTTP auth)

The MIT License (MIT)
Copyright (c) 2016 Baryshnikov Alexander <dev@baryshnikov.net>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import List
from datetime import datetime
from rvcm.cli import *
import re
import json


class Abonent:
    """
    Keep IP and phone number
    """

    def __init__(self, phone: str, ip: str):
        self.phone = phone
        self.ip = ip

    def __repr__(self):
        return self.__class__.__name__ + "(" + ",\n   ".join(k + "=" + repr(v) for k, v in self.__dict__.items()) + ")"

    def __str__(self):
        return repr(self)


class Call:
    """
    Describe single call history record
    """

    def __init__(self, line: int, direction: str, calling: Abonent, called: Abonent, duration_seconds: int,
                 stamp: datetime, status: str):
        self.line = line
        self.direction = direction
        self.calling = calling
        self.called = called
        self.duration = duration_seconds
        self.stamp = stamp
        self.status = status

    def __repr__(self):
        return self.__class__.__name__ + "(" + ",\n   ".join(k + "=" + repr(v) for k, v in self.__dict__.items()) + ")"

    def __str__(self):
        return repr(self)


class History:
    """
    Describes collection of calls history
    """
    URL = '/voice_call_logs.htm?l0=3&l1=2&l2=1&l3=-1'

    def __init__(self, calls: List[Call] = None):
        self.calls = calls or []

    def parse(self, page: str):
        keyword = 'var call_logs ='
        line = '[]'
        for ln in page.splitlines():
            index = ln.find(keyword)
            if index != -1:
                line = ln[index + len(keyword):-1]
                break
        records = json.loads(line)
        # Example:
        # "line 0, Answered, IN, Calling:0000000000000;cpc-rus=1;phone-cont(55.66.77.88),
        # Called:+100000000(11.22.33.44), Duration:0h:15m:34s, Mon Nov 28 19:43:31 2016"
        calls = []
        abonent_pattern = re.compile(r'(?P<phone>[0-9\+\-]+).*\((?P<ip>.*?)\)')
        for record in records:
            line, status, direction, source, target, duration, stamp = map(str.strip, record.split(','))
            line_num = int(line.split()[1])
            calling_phone, calling_ip = abonent_pattern.findall(source.split(':')[1])[0]
            called_phone, called_ip = abonent_pattern.findall(target.split(':')[1])[0]
            _, span = duration.split(':', 1)
            h, m, s = span.split(':')
            seconds = int(h[:-1]) * 3600 + int(m[:-1]) * 60 + int(s[:-1])
            calls.append(Call(
                line=line_num,
                direction=direction.upper(),
                calling=Abonent(phone=calling_phone, ip=calling_ip),
                called=Abonent(phone=called_phone, ip=called_ip),
                duration_seconds=seconds,
                stamp=datetime.strptime(stamp, '%a %b %d %H:%M:%S %Y'),
                status=status
            ))
        self.calls = calls

    def retrieve(self, requester: callable):
        """
        Get information about calls from router
        :param requester: function that returns text by url
        :return: self
        """
        resp = requester(self.URL)
        self.parse(resp.text)
        return self

    def __repr__(self):
        return self.__class__.__name__ + "(" + ",\n   ".join(k + "=" + repr(v) for k, v in self.__dict__.items()) + ")"

    def __str__(self):
        return repr(self)


@cli.group()
def calls():
    """
    Calls operations
    """
    pass


@calls.command()
@click.pass_context
def info(ctx):
    """Print calls history"""
    line = "{line:4}" \
           " {direction:9}" \
           " {status:9}" \
           " {calling_phone:13}" \
           " {calling_ip:16}" \
           " {called_phone:16}" \
           " {called_ip:16}" \
           " {duration:8}" \
           " {stamp:19}"
    print(line.format(
        line='LINE',
        direction='DIRECTION',
        status='STATUS',
        calling_phone='CALLING-PHONE',
        calling_ip='CALLING-IP',
        called_phone='CALLED-PHONE',
        called_ip='CALLING-IP',
        duration='DURATION',
        stamp='STAMP'

    ))
    history = History().retrieve(ctx.obj.getter)
    for call in history.calls:
        print(line.format(
            line=call.line,
            direction=call.direction,
            status=call.status,
            calling_phone=call.calling.phone,
            calling_ip=call.calling.ip,
            called_phone=call.called.phone,
            called_ip=call.called.ip,
            duration=call.duration,
            stamp=call.stamp.isoformat('T')
        ))


@calls.command()
@click.pass_context
def export(ctx):
    """Print calls history in JSON"""
    history = History().retrieve(ctx.obj.getter)
    data = []
    from collections import OrderedDict
    for call in history.calls:
        data.append(OrderedDict([
            ("line", call.line),
            ("direction", call.direction),
            ("status", call.status),
            ("calling_phone", call.calling.phone),
            ("calling_ip", call.calling.ip),
            ("called_phone", call.called.phone),
            ("called_ip", call.called.ip),
            ("duration", call.duration),
            ("stamp", call.stamp.isoformat('T'))
        ]
        ))
    print(json.dumps(data, ensure_ascii=False, indent=4))


if __name__ == '__main__':
    cli(obj=None)
